"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const rx = require("rxjs");
// import * as RxMe from 'rxme';
const RxMe = require("rxme");
exports.State = {
    OK: 'OK',
    ERROR: 'ERROR'
};
class QEntryRun {
    constructor() {
        this.started = new Date();
    }
    error(err) {
        this._error = err;
    }
    completed() {
        this._completed = new Date();
    }
}
exports.QEntryRun = QEntryRun;
class QEntry {
    constructor(input, output) {
        this.runs = [];
        this.input = input;
        this.output = output;
        this.created = new Date();
    }
    isCompleted() {
        return !!this.runs.find(r => !!r.completed);
    }
    isWaiting() {
        // console.log(this.runs);
        return this.runs.length == 0 || !!this.runs.find(qer => !qer.completed);
    }
    newQEntryRun() {
        const qer = new QEntryRun();
        this.runs.push(qer);
        return qer;
    }
}
exports.QEntry = QEntry;
class WorkerState {
    constructor(q, qw) {
        this.q = q;
        this.qWorker = qw;
        this.running = null;
    }
    isFree() {
        return !this.running;
    }
    run(qe) {
        this.running = qe.newQEntryRun();
        this.output = new RxMe.Subject(null);
        this.output.subscribe((a) => {
            // console.log('WorkerState:next', a);
            qe.output.next(a);
        }, (err) => {
            // console.log('WorkerState:error', err);
            this.running.error(err);
            this.running = null;
            qe.output.error(err);
        }, () => {
            // console.log('WorkerState:completed');
            this.running.completed();
            this.running = null;
            qe.output.complete();
            this.q.processMemoryQ();
        });
        this.qWorker(qe.input, this.output);
    }
}
exports.WorkerState = WorkerState;
// export class QueueSubject<T> extends RxMe.Subject<Queue<T>> { }
class Subject extends RxMe.Subject {
    constructor(a) {
        super(a);
    }
}
exports.Subject = Subject;
// export class QueueObservable<T> extends RxMe.Observable<Queue<T>> { }
class Observable extends RxMe.Observable {
}
exports.Observable = Observable;
class Logger {
    constructor(upStream, qid) {
        this.upStream = upStream;
        this.base = [`Q[${qid}]:`];
    }
    info(...args) {
        this.upStream.next(RxMe.logMsg('info', args));
    }
    error(...args) {
        this.upStream.next(RxMe.logMsg('error', this.base.concat(args)));
    }
    debug(...args) {
        this.upStream.next(RxMe.logMsg('debug', this.base.concat(args)));
    }
}
exports.Logger = Logger;
class Queue {
    constructor(obs, argv, id) {
        // public q: Rx.Subject<QEntry<T>>;
        // public readonly deadLetter: Rx.Subject<QEntry<T>>;
        this.qEntries = [];
        // this.inProcess = false;
        // this.obs = obs;
        this.logger = new Logger(obs, id);
        this.lengthSubject = new rx.Subject();
        this.taskTimer = argv.taskTimer || 500;
        this.reclaimTimeout = argv.reclaimTimeout || 10000;
        this.retryWaitTime = argv.retryWaitTime || 1000;
        this.maxExecuteCnt = argv.maxExecuteCnt || 10;
        this.workers = [];
        // this.q = new Rx.Subject<QEntry<T>>();
        // this.q.subscribe(this.action.bind(this));
        // this.deadLetter = new Rx.Subject<QEntry<T>>();
        this.qTask = setInterval(this.processMemoryQ.bind(this), this.taskTimer);
    }
    addWorker(qw) {
        this.workers.push(new WorkerState(this, qw));
        return this;
    }
    length() {
        return this.lengthSubject;
    }
    stop() {
        this.logger.info('run stop');
        // console.log('run stop');
        return RxMe.Observable.create(null, (observer) => {
            this.logger.info('queue stop');
            let forceStopAfter = 3;
            const action = () => {
                if (forceStopAfter <= 0 || this.qEntries.length == 0) {
                    clearInterval(this.qTask);
                    this.logger.info('Q Task stopped');
                    observer.next(RxMe.done(true));
                    observer.complete();
                }
                else {
                    --forceStopAfter;
                    this.logger.info('waiting for Q Task to stop', this.qEntries.length, forceStopAfter);
                    setTimeout(action, this.taskTimer);
                }
            };
            action();
        });
    }
    // private action(qe: QEntry<T>): void {
    //   this.logger.debug('action:', qe);
    //   // qe.running = new Date();
    //   // ++qe.executeCnt;
    //   // qe.task.subscribe((_qe: T) => {
    //   //   qe.running = null;
    //   //   qe.completed = new Date();
    //   //   this.logger.debug('action:completed:', qe);
    //   // }, (error: any) => {
    //   //   qe.running = null;
    //   //   qe.retryAt = new Date((new Date()).getTime() + this.retryWaitTime);
    //   //   this.logger.debug('action:error:', qe.created.getTime(),
    //   //                                      qe.retryAt.getTime(),
    //   //                                      qe);
    //   // });
    // }
    processMemoryQ() {
        const freeWorker = this.workers.find(w => w.isFree());
        const qentry = this.qEntries.find(q => q.isWaiting());
        if (freeWorker && qentry) {
            // this.logger.info('processMemoryQ:run');
            freeWorker.run(qentry);
        }
        const prevLen = this.qEntries.length;
        this.qEntries = this.qEntries.filter(a => !a.isCompleted());
        if (prevLen != this.qEntries.length) {
            this.lengthSubject.next(this.qEntries.length);
        }
        // this.qEntries = this.qEntries.filter((qe) => {
        //   if (qe.completed) {
        //     this.logger.debug('completed remove', qe);
        //     return false;
        //   } else if (qe.executeCnt >= this.maxExecuteCnt) {
        //     this.logger.error('drop queue entry:', qe);
        //     this.deadLetter.next(qe);
        //     return false;
        //   } else if (qe.retryAt) {
        //     if (now >= qe.retryAt.getTime()) {
        //       qe.retryAt = null;
        //       this.logger.debug('retry queue entry:', qe);
        //       this.q.next(qe);
        //     }
        //     return true;
        //   } else if (!qe.running) {
        //     this.logger.debug('trigger action:', qe);
        //     this.q.next(qe);
        //   } else if ((qe.running && qe.executeCnt < this.maxExecuteCnt &&
        //     (now - qe.running.getTime()) >= this.reclaimTimeout)) {
        //     qe.running = null;
        //     this.logger.debug('retry execution time entry:', qe);
        //     this.q.next(qe);
        //   }
        //   return true;
        // });
    }
    push(action, completed) {
        this.qEntries.push(new QEntry(action, completed));
        this.lengthSubject.next(this.qEntries.length);
        this.processMemoryQ();
    }
}
exports.Queue = Queue;
let queueId = ~~(0x1000000 * Math.random());
function start(argv) {
    return Observable.create(Queue, (obs) => {
        const q = RxMe.data(new Queue(obs, argv, queueId++));
        q.asKind().logger.info('Queue Started');
        obs.next(q);
    });
}
exports.start = start;
exports.default = Queue;
//# sourceMappingURL=simqle.js.map